import ply.yacc as yacc
from .latexLex import tokens, latexLexer
from .latexPreprocessing import preprocessImplicitBraces
from sympy import latex

precedence=(
    ('left', 'PLUS'),
    ('nonassoc', 'BIGOP'),
    ('left', 'TIMES'),
)

start = 'expression_list'

def p_ordinary_symbols(p):
    '''expression : PLUS
                  | MINUS
                  | DIVIDE
                  | TIMES
                  | DTIMES
                  | ID
                  | NUMBER
                  | COMMA
                  | EXCL
                  | MOD
                  | EQ
                  | UP LBRACE expression RBRACE
                  '''
    translate = { '\\infty' : 'infty'
                }
    if p[1] == '\\%':
        p[0] = '% '
    elif p[1] == '^':
        p[0] = '^(%s) ' % p[3]
    else:
        p[0] = '%s ' % translate.get(p[1], p[1])
    #print("ordinary_symbols : " + p[0])
                
def p_expression_expr(p):
    '''expression_list : expression'''
    p[0] = p[1]
    #print("expression_expr : "+str(p[0]))
                
def p_expression_list_expr(p):
    '''expression_list : expression_list expression'''
    p[0] = p[1] + p[2]
    #print("expression_list_expr : "+str(p[0]))

def p_expression_parentheses(p):
    '''expression : LPAREN expression_list RPAREN
                  | LFLOOR expression_list RFLOOR
                  | LCEIL expression_list RCEIL
                  | VERT expression_list VERT
                  '''
    tokenToFunc = {'\\lceil' : 'ceil',
                   '\\lfloor' : 'floor',
                   '\\vert' : 'abs',
                   '(' : '',
                  }
    p[0] = '%s(%s)' % (tokenToFunc[p[1]], p[2])
    #print("expression_parentheses : "+str(p[0]))

def p_expression_brackets(p):
    '''expression : LBRACKET expression_list RBRACKET
                  '''
    p[0] = '[%s]' % (p[2])
    #print("expression_brackets : "+str(p[0]))

def p_atom_braces(p):
    '''expression : LBRACE expression_list RBRACE'''
    p[0] = p[2]
    #print("atom_braces : "+str(p[0]))
    
def p_expr_big_op_d(p):
    '''p_expr_big_op_d : BIGOP DOWN LBRACE ID EQ expression RBRACE'''
    p[0] = (p[1], p[4], p[6])
    #print("p_expr_big_op_d : "+str(p[0]))
    
def p_expr_big_op_u(p):
    '''p_expr_big_op_u : BIGOP UP LBRACE expression RBRACE'''
    p[0] = (p[1], p[4])
    #print("p_expr_big_op_u : "+str(p[0]))
    
def p_expr_big_op_du(p):
    '''p_expr_big_op_du : p_expr_big_op_d UP LBRACE expression RBRACE'''
    (operator, index, down) = p[1]
    p[0] = (operator, index, down, p[4])
    #print("p_expr_big_op_du : "+str(p[0]))
    
def p_expr_big_op_ud(p):
    '''p_expr_big_op_ud : p_expr_big_op_u DOWN LBRACE ID EQ expression RBRACE'''
    (operator, up) = p[1]
    p[0] = (operator, p[4], p[6], up)
    #print("p_expr_big_op_ud : "+str(p[0]))
    
def p_expr_big_op(p):
    '''expression : p_expr_big_op_du expression
                  | p_expr_big_op_ud expression'''
                  
    (operator, index, down, up) = p[1]

    tokenToNode = {'\\sum'  : 'sum',
                   '\\prod' : 'prod',
                   '\\int'  : 'integrate',
                  }
    
    p[0] = '%s(%s, %s, %s, %s)'%(tokenToNode[operator], p[2], index, down, up)
    #print("p_expr_big_op : "+str(p[0]))
        
def p_expr_big_opd(p):
    '''expression : BIGOPD DOWN LBRACE ID TO expression_list RBRACE expression_list'''
    tokenToNode = {'\\lim' : 'limit',
                  }
    p[0] = '%s(%s, %s, %s)' % (tokenToNode[p[1]], p[8], p[4], p[6])
    #print("p_expr_big_opd : "+str(p[0]))
        
def p_expr_op_binary(p):
    '''expression : BINARY LBRACE expression_list RBRACE LBRACE expression_list RBRACE'''
    if p[1] == '\\frac':
        p[0] = '((%s)/(%s))' % (p[3], p[6])
    elif p[1] == '\\binom':
       p[0] = 'C(%s, %s)' % (p[3], p[6])
    #print("expr_op_binary : "+str(p[0]))
        
def p_expr_op_unary_optionnal(p):
    '''expression : UNARY LBRACKET expression_list RBRACKET LBRACE expression_list RBRACE'''
    if p[1] == '\\sqrt':
        p[0] = 'root(%s, %s)' % (p[6], p[3])
    #print("expr_op_unary_optionnal : "+str(p[0]))
        
def p_expr_op_unary(p):
    '''expression : UNARY LBRACE expression_list RBRACE'''
    tokenToFunc = {'\\sqrt' : 'sqrt',
                   '\\log' : 'log',
                   '\\exp' : 'exp',
                   '\\cos' : 'cos',
                   '\\sin' : 'sin',
                   '\\tan' : 'tan',
                  }
    p[0] = '%s(%s)' % (tokenToFunc[p[1]], p[3])
    #print("expr_op_unary : "+str(p[0]))

def p_error(p):
    print("Syntax error in input : "+str(p))

latexParser = yacc.yacc(debug=0, write_tables=0)

def latexToCalchas(formula):
    return latexParser.parse(preprocessImplicitBraces(formula), lexer=latexLexer)
    
    



# A lot of this code come from http://docs.sympy.org/dev/_modules/sympy/parsing/mathematica.html
# http://sympy.org/en/index.html

from .mathematicaYacc import mathematicaParser
from .mathematicaLex import mathematicaLexer
from .calchasPreprocessing import calchasToSympy
from .latexPreprocessing import isLatex
from .latexYacc import latexToCalchas
from re import match, sub

class Parser():
    def __init__(self, expr):
        self.expr=expr

    def fromCalchas(self):
        result = calchasToSympy('(%s)' % self.expr)
        if result:
            self.expr=result

    def fromMathematica(self):
        result = mathematicaParser.parse('(%s)' % self.expr, lexer=mathematicaLexer)
        if result:
            self.expr=result.toSympy()

    def fromLatex(self):
        if isLatex(self.expr):
            result = latexToCalchas('(%s)' % self.expr)
            if result:
                self.expr=result

    def normalize(self):
        self.fromLatex()
        self.fromMathematica()
        self.fromCalchas()
        return 'simplify(%s,2)' % self.expr



import re

CARACTERISTIC_SYMBOLS = ['sqrt'] + list(r'*+!\_^[]{}%')
MAYBE_CARACTERISTIC_SYMBOLS = list(r'-/()0123456789=')
ANTI_CARACTERISTIC_SYMBOLS = ['Who', 'who', 'What', 'what', 'Where', 'where', 'Which', 'which', 'How', 'how', 'Why', 'why', 'is', 'Is', 'of'] + list(r'?"éàèùçÉÀÈÙÇâêîôûäëïöüÂÊÎÔÛÄËÏÖÜ')
ANTI_CARACTERISTIC_REGEX = ['sum', 'derivative', 'product', 'limit', 'antiderivative', 'integrate', 'approx', 'approximation']

def isMath(formula):
    isCaract=any(e in formula for e in CARACTERISTIC_SYMBOLS)
    isAnti=any(e in formula for e in ANTI_CARACTERISTIC_SYMBOLS) or any(re.search(r'%s\s+[^\(]'%e, formula)!=None for e in ANTI_CARACTERISTIC_REGEX)
    if isCaract and not isAnti:
        return 2
    if isAnti:
        return 0
    return 1

def isInteresting(inputFormula, outputFormula):
    forbidenChar = list(r' .*+')
    for c in forbidenChar:
        inputFormula = inputFormula.replace(c, "")
    inputFormula = sorted(inputFormula)
    for c in forbidenChar:
        outputFormula = outputFormula.replace(c, "")
    outputFormula = sorted(outputFormula)
    return inputFormula != outputFormula

def traceContainsSpellChecker(trace):
    return any(e.module == 'spell-checker' for e in trace)

def relevance(inputString, outputString):
    lenOutput = len(outputString)
    if lenOutput != 0:
        return len(inputString)/lenOutput
    return 0
